/*
 * This module attempts to provide a more efficient substitute for the
 * rather limited capabilities of the DEC C runtime with regard to
 * process creation/control.  The parent process uses the functions
 * in this module instead fork()/exec() or lib$spawn().  To gain efficiency,
 * we have the spawned processes run a special command procedure as a
 * wrapper that lets us cache the processes and reuse them for several
 * command invocations.
 *
 * At initialization we create 2 mailboxes for process coordination:
 *   - The request mailbox, write only, 512 byte max message.
 *   - The report mailbox, read only 256 byte max message.
 *
 * The unit numbers of these 2 mailboxes are parameters to the wrapper
 * procedure with the third parameter being a timeout value.
 *
 * When we need to execute a command in a sub-process, we write a request
 * message containing the input/output files and the command to execute
 * (creating a process if request mailbox has no current readers).
 * The sub-process executes the command and writes a message to
 * the report mailbox indicating final status (it should assign a channel
 * to the request mailbox prior to sending the report).
 *
 * A process created by this module is always in 1 of 5 states:
 *
 *    0 - Process is starting.
 *    1 - Process ia 'ready' and attempting to read request mailbox.
 *    2 - Process is 'busy', it was the receiver of a request.
 *    3 - Process is 'done' with exit status ready for retrieval by parent.
 *    4 - Process 'dead' process died.
 *
 * Author: David Jones
 * Date:	3-JUL-1999
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stat.h>
#include <fcntl.h>
#include <unixio.h>		/* for isapipe() function */
#include <errno.h>

#include <ssdef.h>		/* VMS system service status return codes */
#include <iodef.h>		/* QIO function codes */
#include <dvidef.h>		/* $GETDVI item codes */
#include <cmbdef.h>		/* Extended mailbox functions */
#include <descrip.h>		/* string descriptors */
#include <msgdef.h>		/* mailbox message codes */
#include <jpidef.h>		/* VMS job/process information codes */
#include <lckdef.h>		/* VMS distributed lock manager codes */

int SYS$CREMBX(), SYS$DASSGN(), SYS$QIOW(), SYS$QIO(), LIB$GETDVI();
int LIB$SPAWN(), LIB$GET_EF(), SYS$DCLAST(), SYS$CANCEL(), SYS$SETAST();
int SYS$HIBER(), SYS$WAKE(), SYS$ENQW(), SYS$DEQ(), LIB$GETJPI();
int SYS$SYNCH();

#include "process_cache.h"	/* verify prototypes */
/*
 * Structure definitions.
 */
struct mbx_iosb {
    unsigned short status, count;
    long pid;
}; 
struct process_info {
    struct process_info *next;
    long pid;				/* process id */
    long lockid;
    int state;				/* 0-new, 1-ready, 2-busy, 3-done */
    int waitable;			/* can return a wait state */
    int last_status;			/* completion status */
};
struct mbx_message {
    unsigned short type;		/* message type */
    unsigned short unit;		/* not used */
    int finalsts;			/* exit status */
    char data[248];
};
/*
 * Global data.  Note that the lists are accessed at AST level.
 */
static char process_startup_cmd[256];
static $DESCRIPTOR(process_startup_dx,process_startup_cmd);
static vms_mailbox_set ctl;
static int report_ef;
static int waiting_for_completion;
static int initial_used = 0;
static struct process_info *free_pcb;	/* lookaside list for procinfo structs*/
static struct process_info *active_pcb;
#define STATIC_PCB_ALLOCATION 128
static struct process_info initial_pcb[STATIC_PCB_ALLOCATION];
/*
 * Create a temporary mailbox with flags and return the channel  unit number.
 * Protection is: SY:RWED, OW:RWED, GR: WO:
 */
static int create_mailbox ( int max_msg, int quota, int flags,
	vms_mailbox *mbx )
{
    int status, code, fd_flags;
    char mbxname[64];
    /*
     * Default structure to NULL setting and test arguments.
     */
    mbx->chan = 0;
    mbx->maxmsg = max_msg;
    mbx->fd = -1;
    mbx->check = 0;			/* don't check for partner */
    mbx->fname = "";
    mbx->bufpos = 0;
    mbx->buffer = (char *) 0;
    if ( max_msg <= 0 ) return 1;	/* OK */
    if ( max_msg > quota ) return SS$_BADPARAM;
    /*
     * Create the mailbox.
     */
    status = SYS$CREMBX (0, &mbx->chan, max_msg, quota, 0x0ff00, 0, 0, flags);
    if ( (status&1) == 0 ) return status;
    /*
     * Lookup unit number.
     */
    code = DVI$_UNIT;
    mbx->unit = 0;
    status = LIB$GETDVI ( &code, &mbx->chan, 0, &mbx->unit, 0, 0 );
    if ( (status&1) == 0 ) {
	/* Cleanup for error */
	mbx->maxmsg = 0;
	SYS$DASSGN ( mbx->chan );
	return status;
    }
    return status;
}
/*
 * Allocation routine.
 */
static struct process_info *new_pcb()
{
    struct process_info *pcb;
    /*
     * Allocate block.
     */
    pcb = free_pcb;
    if ( pcb ) {
	/*
	 * Block available on lookaside list.
	 */
	free_pcb = pcb->next;
    } else if ( initial_used < STATIC_PCB_ALLOCATION ) {
	/*
	 * Block available from static allocation area.
	 */
	pcb = &initial_pcb[initial_used];
	initial_used++;
    } else {
	/*
	 * Take block from heap.
	 */
	pcb = (struct process_info *) malloc ( sizeof(struct process_info) );
    }
    /*
     * Initialize block.
     */
    if ( pcb ) {
	pcb->pid = 0;
	pcb->next = (struct process_info *) 0;
	pcb->state = 0;
	pcb->waitable = 1;
	pcb->last_status = 1;
    }
    return pcb;
}
/***************************************************************************/
/* Define internal routines to handle locks.  Locks are sublocks of
 * resource named "PROCESS_CACHE_xxxxxxxx" when xxxxxxxx is PID of parent
 * process.  The sub-lock resource name is "EXIT_xxxxxxxx" where xxxxxxxx
 * is pid of child process.
 */
static char master_lock_name[32], child_lock_name[32];
static $DESCRIPTOR(master_lock_dx,master_lock_name);
static $DESCRIPTOR(child_lock_dx,"");
static long parent_lock_id;

static void init_lock ( )
{
    int status, code, parent_pid;
    long lksb[2];

    code = JPI$_PID;
    status = LIB$GETJPI ( &code, 0, 0, &parent_pid, 0, 0 );
    sprintf ( master_lock_name, "PROCESS_CACHE_%x", parent_pid );
    master_lock_dx.dsc$w_length = strlen ( master_lock_dx.dsc$a_pointer );

    status = SYS$ENQW ( 0, LCK$K_CRMODE, lksb, 0, &master_lock_dx,
	0, 0, 0, 0, 0, 0 );
    if ( (status&1) == 1 ) status = lksb[0];
    if ( (status&1) == 0 ) fprintf(stderr, "Error getting lock %s: %d\n",
	master_lock_name, status );
    parent_lock_id = lksb[1];
}

static void destroy_lock ( struct process_info *pcb )
{
    SYS$DEQ ( pcb->lockid, 0, 0, 0 );
    pcb->lockid = 0;
}

static void create_lock ( struct process_info *pcb )
{
    char lock_name[32];
    long lksb[2];
    int status;

    sprintf ( lock_name, "EXIT_%x", pcb->pid );
    child_lock_dx.dsc$a_pointer = lock_name;
    child_lock_dx.dsc$w_length = strlen ( child_lock_dx.dsc$a_pointer );

    status = SYS$ENQW ( 0, LCK$K_PWMODE, lksb, 0, &child_lock_dx,
	parent_lock_id, 0, 0, 0, 0, 0 );
    if ( (status&1) == 1 ) status = lksb[0];
    if ( (status&1) == 1 ) pcb->lockid = lksb[1];
    else {
	fprintf(stderr,"Error creating lock %s: %d\n", lock_name, status );
	pcb->lockid = 0;
    }
}

static void convert_lock ( struct process_info *pcb, int new_mode )
{
    int status;
    long lksb[2];

    lksb[1] = pcb->lockid;
    status = SYS$ENQW ( 0, new_mode, lksb, LCK$M_CONVERT, 0, parent_lock_id,
	0, 0, 0, 0, 0 );
    if ( status&1 ) status = lksb[0];
    if ( (status&1) == 0 ) {
	fprintf(stderr, "Error converting lock EXIT_%x (%x): %d\n", 
		pcb->pid, pcb->lockid, status );
    }
}
/***************************************************************************/
static void report_ast ( struct mbx_message *msg )
{
    static struct mbx_message message;
    static struct mbx_iosb iosb;
    int status;

    if ( msg && (iosb.status&1) ) {
	/*
	 * Find sender on active list.
	 */
	struct process_info *pcb;
	for ( pcb = active_pcb; pcb; pcb = pcb->next ) {
	    if ( pcb->pid == iosb.pid ) break;
	}
	if ( pcb ) {
	    /*
	     * Interpret contents of message.
	     */
	    if ( msg->type == SS$_LINKDISCON ) {
		/*
		 * Wrapper process is indicating it is ready for next task.
		 */
		pcb->state = 3;		/* process done */
		if ( waiting_for_completion ) {
			/*
			 * Wake ourselves.
			 */
		    	waiting_for_completion = 0;
			SYS$WAKE(0,0);
		}
	    } else {
	    fprintf(stderr,
		"Invalid message to process cache mbx from PID %x ignored\n",
		iosb.pid );
	    }
	
	} else {
	    fprintf(stderr,
		"Unexpected message to process cache mbx from PID %x ignored\n",
		iosb.pid );
	}
    }
    /*
     * Start new read.
     */
    status = SYS$QIO ( report_ef, ctl.in.chan, IO$_READVBLK, &iosb,
	&report_ast, &message,
	&message, sizeof(message), 0, 0, 0, 0 );
    if ( (status&1) == 0 ) {
	fprintf(stderr,"Process cache QIO failure on report MBX: %d\n",
		status );
    }
}
static void child_death_ast ( struct process_info *pcb )
{
    int state;
    /*
     * Take state-specific actions.
     */
    state = pcb->state;
    pcb->state = 4;
    /* printf("Process %x died, waitable: %d\n", pcb->pid, pcb->waitable ); */
    switch ( state ) {
	case 0:
	case 1:
	    pcb->last_status = SS$_LINKEXIT;
	    break;
	case 2:
	case 3:
	    break;
	case 4:
	default:
	    break;
    }
    /*
     * Make new state '4' and either deallocate block or continue from wait.
     */
    if ( !pcb->waitable && active_pcb ) {
	struct process_info *cur, *prev;
	if ( active_pcb == pcb ) {
	    active_pcb = pcb->next;
	} else {
	    prev = active_pcb;
	    for ( cur = active_pcb->next; cur; cur = cur->next ) {
		if ( cur == pcb ) {
		    prev->next = pcb->next;
		    break;
		}
		prev = cur;
	    }
	    if ( !cur ) return;		/* shouldn't happen */
	}
	destroy_lock ( pcb );
	pcb->next = free_pcb;
	free_pcb = pcb;

    } else if ( waiting_for_completion ) {
	waiting_for_completion = 0;
	SYS$WAKE(0,0);
    }
}
/*
 * Initialize the module.
 */
static int module_init()
{
    int status;
    if ( initial_used > 0 ) return 1;		/* erroneously called */
    /*
     * Create the mailboxes and put he first initial block on the free list.
     */
    free_pcb = initial_pcb;
    free_pcb->next = (struct process_info *) 0;
    active_pcb = (struct process_info *) 0;

    status = vms_create_mailbox_set ( sizeof(struct mbx_message),
	sizeof(struct mbx_message)*2, 512, 512, 0, &ctl );
    if ( (status&1) == 1 ) {
	int dclstatus;
	/*
	 * Close the FDs on the mailbox set so they don't interfere
	 * the the directional features of the mailbox.
	 */
	if ( ctl.in.fd >= 0 ) { close ( ctl.in.fd ); ctl.in.fd = -1; }
	if ( ctl.out.fd >= 0 ) { close ( ctl.out.fd ); ctl.out.fd = -1; }
	report_ef = 0;
	LIB$GET_EF ( &report_ef );
	init_lock();
	/*
	 * Create command that we spawn for the task processes.
	 */
	sprintf ( process_startup_cmd, 
		"@APACHE_CACHED_PROCESS _MBA%d: _MBA%d: %d", ctl.out.unit,
		ctl.in.unit, 600 );
	process_startup_dx.dsc$w_length = strlen(process_startup_cmd);
	/*
	 * Begin AST thread that reads report mailbox.
	 */
	dclstatus = SYS$DCLAST ( report_ast, 0, 0 );
    }
    if ( (status&1) == 1 ) initial_used = 1;
    return status;
}
/***************************************************************************/
/* AST to handle asynchronous reads of the error mailbox.
 */
static void async_err_ast ( vms_mailbox_set *set )
{
    vms_async_buf *tail;
    int status;
    /*
     * Set the length for the current tail of the chain.
     */
    set->last->length = set->err_iosb.status;
    if ( (set->err_iosb.status&1) == 1 ) {
	set->last->length = set->err_iosb.count;
    } else {
	if ( set->err_iosb.status == SS$_NOWRITER ) set->async_err_active = 0;
	set->last->length = -1;
    }
    if ( set->async_err_active == 0 ) return;	/* AST thread cancelled */
    /*
     * Append a new buffer to the chain and issue another read.
     */
    tail = (vms_async_buf *) malloc ( sizeof(vms_async_buf) +
	set->err.maxmsg );
    if ( !tail ) return;
    tail->next = (vms_async_buf *) 0;
    tail->status = SS$_ABORT;
    tail->length = -1;
    set->last->next = tail;
    set->last = tail;

    status = SYS$QIO ( 0, set->err.chan, IO$_READVBLK|IO$M_WRITERCHECK, 
	&set->err_iosb, async_err_ast, set,
	&set->last->data[0], set->err.maxmsg, 0, 0, 0, 0 );

    if ( (status&1) == 0 ) {
	tail->status = status;
	set->err_iosb.status = status;
    }
}
static vms_async_buf *dequeue_async_buf ( vms_mailbox_set *set, int keep_active )
{
    vms_async_buf *head;
    SYS$SETAST(0);
    head = set->first;
    if ( head && set->async_err_active ) {
	if ( !keep_active ) {
	    /*
	     * Cancel any pending read and wait for cancel to complete.
	     */
	    SYS$CANCEL ( set->err.chan );
	    set->async_err_active = 0;
	    SYS$SETAST(1);
	    SYS$SYNCH ( 0, &set->err_iosb );
	    SYS$SETAST(0);
	} else {
	    /*
	     * Don't return the last buffer, instead wait for completion
	     * of next buffer that lets AST routine add to chain.
	     */
	    if ( head == set->last && !set->err.check ) {
		int i;
		for ( i = 0; (head == set->last) && i < 500; i++ ) {
	    	    SYS$SETAST(1);
		    SYS$SYNCH ( 0, &set->err_iosb );
		    SYS$SETAST(0);
		}
		if ( head == set->last ) head = (vms_async_buf *) 0;
	    }
	}
    }
    if ( head ) set->first = head->next;
    SYS$SETAST(1);
    return head;
}
/***************************************************************************/
/* Create 3 directional mailboxes and use to initialize a mailbox set
 * structure.
 */
int vms_create_mailbox_set ( int in_maxmsg, int in_quota,
	int out_maxmsg, int out_quota, int err_maxmsg, vms_mailbox_set *set )
{
    int status;

    set->task = 0;
    set->in.maxmsg = 0;
    set->out.maxmsg = 0;
    set->out.maxmsg = 0;
    set->async_err_active = 0;
    set->first = set->last = (vms_async_buf *) 0;
    /*
     * Create the 3 mailboxes.
     */
    status = create_mailbox(in_maxmsg, in_quota, CMB$M_READONLY, &set->in);
    if ( (status&1) == 1 ) {
	status = create_mailbox ( out_maxmsg, out_quota, CMB$M_WRITEONLY,
		&set->out );
	if ( (status&1) == 1 ) {
	    status = create_mailbox ( err_maxmsg, err_maxmsg, CMB$M_READONLY,
		&set->err );
	    if ( ((status&1) == 1) && (err_maxmsg > 0) ) {
		/*
		 * Initiate asyncrhonous reads on error mailbox.
		 */
		set->async_err_active = 1;
		set->first = (vms_async_buf *) malloc (
			sizeof(vms_async_buf) + err_maxmsg );
		set->last = set->first;
		set->first->next = (vms_async_buf *) 0;
		set->first->length = -1;

		status = SYS$QIO( 0, set->err.chan, IO$_READVBLK,
			&set->err_iosb, async_err_ast, set,
			&set->first->data[0], err_maxmsg, 0, 0, 0, 0 );
	    }
	}
    }
    if ( (status&1) == 0 ) {
	/*
	 * Error, rollback any mailboxes that were created.
	 */
    }
    return status;
}
/***************************************************************************/
/* Main routine for creating new process.  Send message to request mailbox.
 */
int vms_create_task ( vms_mailbox_set *set, char *command,  char **env,
	long *pid )
{
    char request[512];
    int i, status, req_len, flags;
    char *input, *output, *error, *envmbx;
    char input_fname[32], output_fname[32], error_fname[32];
    struct mbx_iosb iosb;
    struct process_info *pcb;
    /*
     * Build request line, delimit with NULs (allows spaces in filenames).
     */
    status = module_init();
    if ( (status&1) == 0 ) return status;

    if ( set->in.maxmsg > 0 ) {
	input = input_fname;
	sprintf ( input_fname, "_MBA%d:", set->in.unit );
    } else input = set->in.fname;

    if ( set->out.maxmsg > 0 ) {
	output = output_fname;
	sprintf ( output_fname, "_MBA%d:", set->out.unit );
    } else {
	env = (char **) 0;	/* Can't transfer env if not a mailbox */
	output = set->out.fname;
    }

    if ( set->err.maxmsg > 0 ) {
	error = error_fname;
	sprintf ( error_fname, "_MBA%d:", set->err.unit );
    } else error = set->err.fname;

    envmbx = "";
    if ( env ) envmbx = output;
    req_len = strlen(input)+strlen(output)+strlen(error)+strlen(command)
	+ strlen(envmbx) + 5;
    if ( req_len > sizeof(request) ) return SS$_DATAOVERUN;
    sprintf ( request, "%s%c%s%c%s%c%s%c%s%c", input, 0, output, 0,
	error, 0, command, 0, envmbx, 0 );
    /*
     * Attempt to send to an already-spawned process.
     */
    status = SYS$QIOW ( 0, ctl.out.chan, IO$_WRITEVBLK|IO$M_READERCHECK,
	&iosb, 0, 0, request, req_len, 0, 0, 0, 0 );
    if ( (status&1) == 1 ) status = iosb.status;
    if ( status == SS$_NOREADER ) {
	/*
	 * No process available, spawn a new one.
	 */
	SYS$SETAST(0);
	pcb = new_pcb();
	if ( !pcb ) {
	    SYS$SETAST(1);
	    return SS$_INSFMEM;
	}
	pcb->next = active_pcb;
	active_pcb = pcb;
	SYS$SETAST(1);
	flags = 1;			/* nowait */
	status = LIB$SPAWN ( &process_startup_dx, 0, 0, &flags, 0,
		&pcb->pid, &pcb->last_status, &report_ef, child_death_ast,
		pcb, 0, 0, 0 );
	if ( (status&1) == 1 ) {
	    /*
	     * Spawn successfull, wait on it to initialize.
	     */
	    int estatus, ipid;
	    create_lock ( pcb );
	    ipid = vms_wait_for_task_exit ( pcb->pid, &estatus, 0 );
	    if ( (ipid < 0) || ((estatus&1) == 0) ) {
		fprintf(stderr,"new process died\n");
		return SS$_LINKEXIT;
	    }
	    /*
	     * Retry write to request mailbox, but this time wait.
	     * Note that process that reads the message this time may not
	     * be the same one we just created.
	     */
	    status = SYS$QIOW ( 0, ctl.out.chan, IO$_WRITEVBLK,
		&iosb, 0, 0, request, req_len, 0, 0, 0, 0 );
	    if ( (status&1) == 1 ) status = iosb.status;
	    *pid = iosb.pid;
	    set->task = iosb.pid;
	
	} else {
	    /*
	     * cleanup after error, assume pcb still at head of list.
	     */
	    SYS$SETAST(0);
	    active_pcb = pcb->next;
	    pcb->next = free_pcb;
	    free_pcb = pcb;
	    SYS$SETAST(1);
	}
    }
    if ( (status&1) == 1 ) {
	/*
	 * A process was waiting or created, Lookup reader in list.
	 */
	SYS$SETAST(0);
	for ( pcb = active_pcb; pcb; pcb = pcb->next ) {
	    if ( pcb->pid == iosb.pid ) break;
	}
	if ( pcb ) {
	    *pid = iosb.pid;
	    set->task = iosb.pid;
	    pcb->waitable = 1;				/* process sent req. */
	    if ( pcb->state < 2 ) pcb->state = 2;		/* mark busy */
	} else {
	    SYS$SETAST(1);
	    return SS$_ABORT;		/* couldn't find process in list */
	}
	SYS$SETAST(1);
	/*
	 * Get lock that keeps child from continuing until previous exit
	 * acknowleged.
	 */
	convert_lock ( pcb, LCK$K_PWMODE );
	/*
	 * Send environment array if present.
	 */
	if ( env ) {
	    for ( i = 0; (status&1) && env[i]; i++ ) {
	        status = SYS$QIOW ( 0, set->out.chan, 
			IO$_WRITEVBLK|(set->out.check?IO$M_READERCHECK:0),
			&iosb, 0, 0, env[i], strlen(env[i]), 0, 0, 0, 0 );
		set->out.check = 1;
	        if ( (status&1) == 1 ) status = iosb.status;
	    }
	    if ( (status&1) ) {
	        status = SYS$QIOW ( 0, set->out.chan, 
		    IO$_WRITEOF|(set->out.check?IO$M_READERCHECK:0),
			&iosb, 0, 0, 0, 0, 0, 0, 0, 0 );
	        if ( (status&1) == 1 ) status = iosb.status;
	    }
	}
    } else {
	/*
	 * Unexpected error.
	 */
    }
    return status;
}
/**************************************************************************/
/* Indicate that process will not be waited on.
 */
int vms_detach_task ( int ipid )
{
    /*
     * Check for empty list.
     */
    struct process_info *pcb, *prev, copy;
    int status, waitable_count;
    long pid;

    status = module_init();
    if ( (status&1) == 0 ) {
	return -1;
    }
    /*
     * Scan list looking for a ready or dead pcb.
     */
    status = SS$_NONEXPR;		/* default to error status */
    pid = ipid;
    SYS$SETAST(0);
    if ( active_pcb ) {
	/*
	 * Look for process to detach.
	 */
	prev = (struct process_info *) 0;
	waitable_count = 0;
	for ( pcb = active_pcb; pcb; pcb = pcb->next ) {
	    if ( (pid == pcb->pid) && pcb->waitable ) {
		/*
		 * Found process mark as not waitable.
		 */
		pcb->waitable = 0;
		convert_lock ( pcb, LCK$K_NLMODE );
		status = 1;
		if ( pcb->state == 4 ) {
		    destroy_lock ( pcb );
		    if ( prev ) prev->next = pcb->next;
		    else active_pcb = pcb->next;
		    pcb->next = free_pcb;
		    pcb->state = -1;
		    free_pcb = pcb;
	    	}
		break;
	    }
	    prev = pcb;
	}
    }
    SYS$SETAST(1);
    return status;
}
/**************************************************************************/
/* Wait for specific process to complete.  Flags may contain bit WNOHANG.
 */
int vms_wait_for_task_exit ( int ipid, int *estatus, int flags )
{
    /*
     * Check for empty list.
     */
    struct process_info *pcb, *prev, copy;
    int status, waitable_count;
    long pid;

    status = module_init();
    if ( (status&1) == 0 ) {
	*estatus = status;
	return -1;
    }
    /*
     * Scan list looking for a ready or dead pcb.
     */
    pid = ipid;
    SYS$SETAST(0);
    copy.pid = -1;			/* default to error status */
    while ( active_pcb ) {
	/*
	 * Look for completed (non-busy), starting (state=0) or dead task.
	 */
	prev = (struct process_info *) 0;
	waitable_count = 0;
	for ( pcb = active_pcb; pcb; pcb = pcb->next ) {
	    if ( ((pid == -1) || (pid == pcb->pid)) && pcb->waitable ) {
		waitable_count++;
		if ( (pcb->state > 2) ) {
		    /*
		     * We have a match, free the block if dead.
		     */
		    if ( pcb->state == 4 ) {
			destroy_lock ( pcb );
			if ( prev ) prev->next = pcb->next;
			else active_pcb = pcb->next;
			pcb->next = free_pcb;
			pcb->state = -1;
			free_pcb = pcb;
	    	    } else if ( pcb->state == 3 ) {
			/* reset to 'ready' state */
			pcb->state = 1;
			pcb->waitable = 0;
		    } else {
			/*
			 * 
			 */
		    }
		    break;
		}
	    }
	    prev = pcb;
	}
	/*
	 * Check scan result.
	 */
	if ( waitable_count == 0 ) {
	    /* 
	     * Wait can never be satisfied (all processes have already
	     * returned a completion status)
	     */
	    copy.pid = -1;
	    break;
	} else if ( !pcb && ((flags&WNOHANG) == 0) ) {
	    /*
	     * No match, stall.
	     */
	    waiting_for_completion = 1;
	    SYS$SETAST(1);
	    SYS$HIBER();
	    SYS$SETAST(0);
	} else if ( !pcb ) {
	    /* Nohang flag set, return 0 */
	    copy.pid = 0;
	    copy.last_status = 0;
	    break;
	} else {
	    copy = *pcb;
	    break;
	}
    }
    SYS$SETAST(1);
    /*
     * At loop exit, copy has copy of process_info struct that completed
     * or has 0/-1 as the PID.
     */
    if ( (copy.pid > 0) && (copy.state != -1) ) 
	convert_lock ( &copy, LCK$K_NLMODE );
    *estatus = copy.last_status;
    return (int) copy.pid;
}
/***********************************************************************/
/*
 * Return file descriptor.
 */
int vms_open_mailbox ( vms_mailbox *mbx, int flags )
{
    int status;
    char *fname, mbxname[32];

    if ( mbx->maxmsg > 0 ) {
	fname = mbxname;
	sprintf ( mbxname, "_MBA%d:", mbx->unit );
    } else fname = mbx->fname;

    mbx->fd = open ( fname, flags, 0700 );
    return (mbx->fd < 0) ? 0 : 1;
}
/**************************************************************************/
/* Emulate the apache ap_bgets function.
 *
 * Reads from the stream into the array pointed to by buff, until
 * a (CR)LF sequence is read, or end-of-file condition is encountered
 * or until n-1 bytes have been stored in buff. If a CRLF sequence is
 * read, it is replaced by a newline character.  The string is then
 * terminated with a null character.
 *
 * Returns the number of bytes stored in buff (not including the terminating
 * NULL), or zero on end of transmission, or -1 on an error.
 *
 * Notes:
 *  If null characters are expected in the data stream, then
 * buff should not be treated as a null terminated C string; instead
 * the returned count should be used to determine the length of the
 * string.
 *  CR characters in the byte stream not immediately followed by a LF
 * will be preserved.
 */
int vms_mailbox_gets ( char *buf, int bufsize, vms_mailbox *in )
{
    int status, count, i;
    struct { unsigned short status, count; long pid; } iosb;
    char c;

    if ( in->maxmsg <= 0 ) return -1;	/* not a valid mailbox */
    if ( in->buffer == (char *) 0 ) {
	/*
	 * allocate a buffer, include room for LF+NULL to be tacked to end.
	 */
	in->buffer = malloc ( in->maxmsg+2 );
	if ( !in->buffer ) return -1;
	in->bufpos = 0;		/* next element to read */
	in->buflen = 0;
    }

    if ( bufsize > in->maxmsg ) bufsize = in->maxmsg+1;
    for ( count = 0; ; ) {
	if ( in->bufpos >= in->buflen ) {
	    /*
	     * read more data.  Check for no writer only after first read.
	     */
	    status = SYS$QIOW ( 0, in->chan, 
	        IO$_READVBLK|(in->check?IO$M_WRITERCHECK:0),
	        &iosb, 0, 0,
	        in->buffer, in->maxmsg, 0, 0, 0, 0 );

	    in->check = 1;
	    if ( (status&1) == 1 ) status = iosb.status;
	    if ( (status&1) == 1 ) {
	        in->buflen = iosb.count;
		in->buffer[in->buflen++] = '\n';
		in->buffer[in->buflen] = '\0';
		in->bufpos = 0;
	    } else if ( status == SS$_NOWRITER ) {
		free ( in->buffer );
		in->buffer = (char *) 0;
	        break;
	    } else if ( status == SS$_ENDOFFILE ) {
	        continue;
	    } else {
	        count = -1;
	        errno = EVMSERR;
	        vaxc$errno = status;
	        return count;
	    }
	}
	/*
	 * Move characters into buffer.
	 */
	c = '\0';
	for ( i = in->bufpos; (c!='\n') && (i < in->buflen); i++ ) {

	    if ( count >= (bufsize-1) ) break;
	    c = in->buffer[i];
	    if ( c == '\r' ) {
		/* since LF was appended we know buffer[i+1] exists. */
		if ( in->buffer[i+1] == '\n' ) buf[count++] = c;
	    } else {
		buf[count++] = c;
	    }
	}
	in->bufpos = i;
	if ( c == '\n' ) break;
    }
    /*
     * Always terminate the string (differs from fgets() function).
     */
    buf[count] = '\0';
    return count;
}
/*
 * Duplicate gets functionality only with err mailbox.
 */
int vms_err_mailbox_gets ( char *buf, int bufsize, vms_mailbox_set *set )
{
    int status, count, i;
    vms_mailbox *in;
    struct { unsigned short status, count; long pid; } iosb;
    char c;

    in = &set->err;
    if ( set->err.maxmsg <= 0 ) return -1;	/* not a valid mailbox */
    if ( in->buffer == (char *) 0 ) {
	/*
	 * allocate a buffer, include room for LF+NULL to be tacked to end.
	 */
	in->buffer = malloc ( in->maxmsg+2 );
	if ( !in->buffer ) return -1;
	in->bufpos = 0;		/* next element to read */
	in->buflen = 0;
    }

    if ( bufsize > in->maxmsg ) bufsize = in->maxmsg+1;
    for ( count = 0; ; ) {
	if ( in->bufpos >= in->buflen ) {
	    /*
	     * read more data, pull from head of async queue.
	     */
	    vms_async_buf *cur; int length;

	    cur = dequeue_async_buf ( set, 1 );
	    if ( cur ) { 
		/*
		 * Copy contents of buffer and free async buffer.
		 */
		status = cur->status; 
		length = cur->length; 
		memcpy ( in->buffer, cur->data, length );
		free ( cur );
	    }
	    else { status = SS$_ABORT; length = 0; }
	    if ( (status&1) == 1 ) {
	        in->buflen = length;
		in->buffer[in->buflen++] = '\n';
		in->buffer[in->buflen] = '\0';
		in->bufpos = 0;
	    } else if ( status == SS$_NOWRITER ) {
		free ( in->buffer );
		in->buffer = (char *) 0;
	        break;
	    } else if ( status == SS$_ENDOFFILE ) {
	        continue;
	    } else {
	        count = -1;
	        errno = EVMSERR;
	        vaxc$errno = status;
	        return count;
	    }
	}
	/*
	 * Move characters into buffer.
	 */
	c = '\0';
	for ( i = in->bufpos; (c!='\n') && (i < in->buflen); i++ ) {

	    if ( count >= (bufsize-1) ) break;
	    c = in->buffer[i];
	    if ( c == '\r' ) {
		/* since LF was appended we know buffer[i+1] exists. */
		if ( in->buffer[i+1] == '\n' ) buf[count++] = c;
	    } else {
		buf[count++] = c;
	    }
	}
	in->bufpos = i;
	if ( c == '\n' ) break;
    }
    /*
     * Always terminate the string (differs from fgets() function).
     */
    buf[count] = '\0';
    return count;
}
/*
 * Return pointer to remaining buffer.
 */
int vms_mailbox_buffer ( vms_mailbox *in, char **buffer, int *buflen )
{
    if ( in->maxmsg <= 0 ) return SS$_BADPARAM;
    if ( !in->buffer ) return SS$_ENDOFFILE;
    *buflen = in->buflen - in->bufpos;
    *buffer = &in->buffer[in->bufpos];
    in->bufpos = in->buflen;		/* only return data once */
    return 1;
}
/*************************************************************************/
/*  Deassign mailbox channels in mailbox set.
 */
int vms_destroy_mailbox_set ( vms_mailbox_set *set )
{
    if ( set->in.maxmsg > 0 ) {
	SYS$DASSGN ( set->in.chan );
        if ( set->in.buffer ) free ( set->in.buffer );
    }
    if ( set->out.maxmsg > 0 ) SYS$DASSGN ( set->out.chan );
    if ( set->err.maxmsg > 0 ) {
	/*
	 * Drain any accumated async message buffers.
	 */
	vms_async_buf *cur;

	while ( cur = dequeue_async_buf ( set, 0 ) ) free ( cur );
	SYS$DASSGN ( set->err.chan );
	if ( set->err.buffer ) free ( set->err.buffer );
    }

    set->in.maxmsg = set->out.maxmsg = set->err.maxmsg = 0;
    return 1;
}
